from __future__ import annotations

from granular_configuration_language.yaml.decorators import (
    Root,
    Tag,
    as_lazy_with_root,
    interpolate_value_with_ref,
    string_tag,
)
from granular_configuration_language.yaml.decorators.interpolate import interpolation_needs_ref_condition


@string_tag(Tag("!Sub"), "Formatter")
@as_lazy_with_root(needs_root_condition=interpolation_needs_ref_condition)
@interpolate_value_with_ref
def handler(value: str, root: Root) -> str:
    return value



from __future__ import annotations

import abc
import collections.abc as tabc
import dataclasses
import sys
import typing as typ
from functools import wraps

from ruamel.yaml import MappingNode, Node, SafeConstructor, ScalarNode, SequenceNode

from granular_configuration_language.exceptions import ErrorWhileLoadingTags
from granular_configuration_language.yaml.classes import RT, StateHolder, T, Tag
from granular_configuration_language.yaml.load._constructors import construct_mapping, construct_sequence

if sys.version_info >= (3, 12):
    from typing import override
elif typ.TYPE_CHECKING:
    from typing_extensions import override
else:

    def override(func: typ.Callable) -> typ.Callable:
        return func


Category = typ.NewType("Category", str)
SortedAs = typ.NewType("SortedAs", str)
FriendlyType = typ.NewType("FriendlyType", str)


@dataclasses.dataclass(frozen=True, eq=False, slots=True, repr=False)
class TagConstructor:
    """
    Links the YAML Constructor to the Tag Logic
    """

    tag: Tag
    category: Category
    sort_as: SortedAs
    friendly_type: FriendlyType
    constructor: tabc.Callable[[type[SafeConstructor], StateHolder], None]
    plugin: str = dataclasses.field(default="Unknown", init=False)

    def set_plugin(self, plugin: str) -> None:
        object.__setattr__(self, "plugin", plugin)

    @override
    def __eq__(self, value: object) -> bool:
        return (isinstance(value, self.__class__) and self.tag == value.tag) or (
            isinstance(value, str) and self.tag == value
        )

    @override
    def __hash__(self) -> int:
        return hash(self.tag)

    def __call__(self, constructor: type[SafeConstructor], state: StateHolder) -> None:
        return self.constructor(constructor, state)

    @override
    def __repr__(self) -> str:
        return f"<TagConstructor(`{self.tag}`): {self.constructor.__module__}.{self.constructor.__name__}>"


class TagDecoratorBase(typ.Generic[T], abc.ABC):
    """Base class for Tag Decorator factories.

    You must implement the :py:attr:`user_friendly_type` property and define the generic type.

    Example:

    .. code-block:: python

        class string_tag(TagDecoratorBase[str]):
            Type: typing.TypeAlias = str

            @property
            def user_friendly_type(self) -> str:
                return "str"

    You must override at least one of :py:meth:`scalar_node_type_check`, :py:meth:`sequence_node_type_check`,or
    :py:meth:`mapping_node_type_check`.

    - For :py:meth:`scalar_node_type_check` to be called the YAML has already be tested to be a :py:class:`str`.
    - For :py:meth:`sequence_node_type_check` to be called the YAML has already be tested to be a
      :py:class:`~collections.abc.Sequence`.
    - For :py:meth:`mapping_node_type_check` to be called the YAML has already be tested to be a
      :py:class:`~collections.abc.Mapping`.

    If these are enough, then you may just return :py:data:`True` in the override method. Otherwise, implement the
    override as a :py:data:`~typing.TypeGuard`.

    If the value needs to be altered before being passed to Tag functions, override :py:meth:`scalar_node_transformer`,
    :py:meth:`scalar_node_transformer`, or :py:meth:`mapping_node_transformer`, as needed.

    The transformer is called if the associated node type check passes, just before the value is passed to tag function.
    """

    __slots__ = ("tag", "category", "sort_as")

    def __init__(self, tag: Tag, category: str = "General", *, sort_as: str | None = None) -> None:
        """
        :param Tag tag:
            Value of Tag.
            Expected to be constructed inline using :py:class:`.Tag` (e.g. ``Tag("!Tag")``).
            Must start with ``!``.
        :param str, optional category:
            Category of Tag. Used by :ref:`available_tags <available_tags>` and :ref:`available_plugins <available_plugins>` to organize tags, defaults to ``General``.
        :param str, optional sort_as:
            Alternative Tag string. Used for sorting tags different to its explict value. Used by :ref:`available_tags <available_tags>` and :ref:`available_plugins <available_plugins>`.
        """
        self.tag: typ.Final = tag
        self.category: typ.Final = Category(category)
        self.sort_as: typ.Final = SortedAs(sort_as or tag)

        if not tag.startswith("!"):
            raise ErrorWhileLoadingTags(f"Tag `{tag}` error: All tags must begin with `!`.")

    @property
    @abc.abstractmethod
    def user_friendly_type(self) -> str:
        """User-friendly of the type expected by Tag Decorator.

        Note:
            - Use Python types for consistent communication.
            - This is used when generating exception messages.

        :returns: User-friendly string
        :rtype: str

        """
        ...

    def scalar_node_type_check(self, value: str) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Scalar Node support.

        Parameters:
            value (str): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def sequence_node_type_check(self, value: tabc.Sequence) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Sequence Node support.

        Parameters:
            value (~collections.abc.Sequence): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def mapping_node_type_check(self, value: tabc.Mapping) -> typ.TypeGuard[T]:
        """Defaults to :py:data:`False`. Override to enable Mapping Node support.

        Parameters:
            value (~collections.abc.Mapping): YAML value

        Returns:
            ~typing.TypeGuard[T]: Return :py:data:`True`, if ``value`` is supported.

        """
        return False

    def scalar_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`scalar_node_type_check` return :py:data:`True`.

        As an example, a float tag could be supported by:

        .. code-block:: python

            def scalar_node_type_check(self, value: Any) -> TypeGuard[float]:
                try:
                    float(value)
                    return True
                except ValueError:
                    return False


            def scalar_node_transformer(self, value: Any) -> float:
                return float(value)

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def sequence_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`sequence_node_type_check` return :py:data:`True`.

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def mapping_node_transformer(self, value: typ.Any) -> T:
        """Defaults to an identity operation. Override if the value needs to be altered before being passed to Tag functions.

        Only called if :py:meth:`mapping_node_type_check` return :py:data:`True`.

        Parameters:
            value (~typing.Any): YAML value

        Returns:
            T: Transformed value

        """
        return value

    def __call__(self, handler: tabc.Callable[[Tag, T, StateHolder], RT]) -> TagConstructor:
        # """Takes the wrapped tag function and further wraps it for configuration loading.
        # :param (~collections.abc.Callable[[Tag, T, StateHolder], RT]) handler: Wrapped Tag Function
        # :return: Tag Function ready to be used when loading configuration
        # :rtype: TagConstructor
        # :meta private:
        # """
        # autodoc refuses to exclude `__call__`, even though this is not a part of the public interface

        # Don't capture self in the function generation
        tag = self.tag
        user_friendly_type = FriendlyType(self.user_friendly_type)
        category = self.category
        sort_as = self.sort_as
        scalar_node_type_check = self.scalar_node_type_check
        sequence_node_type_check = self.sequence_node_type_check
        mapping_node_type_check = self.mapping_node_type_check
        scalar_node_transformer = self.scalar_node_transformer
        sequence_node_transformers = self.sequence_node_transformer
        mapping_node_transformer = self.mapping_node_transformer

        @wraps(handler)
        def add_handler(
            constructor: type[SafeConstructor],
            state: StateHolder,
        ) -> None:
            @wraps(handler)
            def type_handler(constructor: SafeConstructor, node: Node) -> RT:
                if isinstance(node, ScalarNode):
                    value = constructor.construct_scalar(node)
                    if isinstance(value, str) and scalar_node_type_check(value):
                        return handler(tag, scalar_node_transformer(value), state)
                elif isinstance(node, SequenceNode):
                    value = construct_sequence(state.options.sequence_func, constructor, node)
                    if isinstance(value, tabc.Sequence) and sequence_node_type_check(value):
                        return handler(tag, sequence_node_transformers(value), state)
                elif isinstance(node, MappingNode):
                    value = construct_mapping(state.options.obj_pairs_func, constructor, node)
                    if isinstance(value, tabc.Mapping) and mapping_node_type_check(value):
                        return handler(tag, mapping_node_transformer(value), state)
                else:
                    pass  # pragma: no cover

                # Fallback Exception
                raise ValueError(f"`{tag}` supports: {user_friendly_type}. Got: `{repr(node)}`")

            constructor.add_constructor(tag, type_handler)

        return TagConstructor(tag, category, sort_as, user_friendly_type, add_handler)



from __future__ import annotations

import collections.abc as tabc
import sys
import typing as typ
from collections import OrderedDict

from granular_configuration_language.exceptions import ErrorWhileLoadingTags
from granular_configuration_language.yaml.decorators._base import Tag, TagConstructor

if sys.version_info >= (3, 12):
    from typing import override
elif typ.TYPE_CHECKING:
    from typing_extensions import override
else:

    def override(func: typ.Callable) -> typ.Callable:
        return func


class TagSet(tabc.Iterable[TagConstructor], tabc.Container[str]):
    def __init__(self, tags: tabc.Iterable[TagConstructor]) -> None:
        self.__state: OrderedDict[Tag, TagConstructor] = OrderedDict()

        for tc in tags:
            tag = tc.tag
            if tag in self.__state:
                raise ErrorWhileLoadingTags(
                    f"Tag is already defined. `{repr(tc)}` attempted to replace `{repr(self.__state[tag])}`"
                )
            else:
                self.__state[tag] = tc

    @override
    def __contains__(self, x: typ.Any) -> bool:
        return x in self.__state

    @override
    def __iter__(self) -> tabc.Iterator[TagConstructor]:
        return iter(self.__state.values())

    @override
    def __repr__(self) -> str:
        return f"TagSet{{{','.join(sorted(self.__state.keys()))}}}"

    def has_tags(self, *tags: Tag | str) -> bool:
        return all(map(self.__contains__, tags))

    def does_not_have_tags(self, *tags: Tag | str) -> bool:
        return not any(map(self.__contains__, tags))

    def get_subset(self, *select: str) -> TagSet:
        def subset() -> tabc.Iterator[TagConstructor]:
            for tag in self:
                if tag.tag in select:
                    yield tag

        return TagSet(subset())



from __future__ import annotations

import collections.abc as tabc
import sys
import typing as typ

from granular_configuration_language.yaml.classes import RT, LazyEval, LazyRoot, Root, Tag

if sys.version_info >= (3, 12):
    from typing import override
elif typ.TYPE_CHECKING:
    from typing_extensions import override
else:

    def override(func: typ.Callable) -> typ.Callable:
        return func


class LazyEvalBasic(LazyEval[RT]):
    def __init__(self, tag: Tag, value: tabc.Callable[[], RT]) -> None:
        super().__init__(tag)
        self.__value = value

    @override
    def _run(self) -> RT:
        return self.__value()


class LazyEvalWithRoot(LazyEval[RT]):
    def __init__(self, tag: Tag, root: LazyRoot, value: tabc.Callable[[Root], RT]) -> None:
        super().__init__(tag)
        self.__value = value
        self.__lazy_root = root

    @override
    def _run(self) -> RT:
        return self.__value(self.__lazy_root.root)



# isort:skip_file
if __name__ == "__main__":
    import argparse
    import operator as op

    from granular_configuration_language.yaml._tags import handlers
    from granular_configuration_language.yaml.decorators._viewer import AvailableTags, can_table

    choices = ["csv", "json"]
    default = "csv"

    if can_table:
        choices.append("table")
        default = "table"

    parser = argparse.ArgumentParser(
        description="Shows available tags", epilog=AvailableTags(handlers).table(_force_missing=True)
    )
    parser.add_argument("type", default=default, choices=choices, nargs="?")

    args = parser.parse_args()

    print(op.methodcaller(args.type)(AvailableTags(handlers)))



from __future__ import annotations

import collections.abc as tabc
import inspect
import os
import typing as typ
from importlib import import_module
from importlib.metadata import entry_points
from importlib.util import resolve_name
from itertools import chain, filterfalse, starmap
from operator import attrgetter, itemgetter
from pathlib import Path

from granular_configuration_language.yaml.decorators._base import Tag, TagConstructor
from granular_configuration_language.yaml.decorators._tag_set import TagSet

ModuleName = typ.NewType("ModuleName", str)
PluginName = typ.NewType("PluginName", str)


def is_TagConstructor(obj: typ.Any) -> typ.TypeGuard[TagConstructor]:
    return isinstance(obj, TagConstructor)


def get_tags_in_module(module_name: ModuleName, plugin_name: PluginName) -> tabc.Iterator[TagConstructor]:
    def set_plugin(tc: TagConstructor) -> TagConstructor:
        tc.set_plugin(plugin_name)
        return tc

    return map(set_plugin, map(itemgetter(1), inspect.getmembers(import_module(module_name), is_TagConstructor)))


def get_internal_tag_plugins() -> tabc.Iterator[ModuleName]:
    import granular_configuration_language.yaml._tags as tags

    tags_package = tags.__package__
    tags_module_path = Path(tags.__file__).parent
    private_sub_module_pattern = r"_[a-zA-Z]*.py"

    def get_abs_name(module_name: str) -> ModuleName:
        return ModuleName(resolve_name("." + module_name, package=tags_package))

    return map(
        get_abs_name,
        filter(
            None,
            map(
                inspect.getmodulename,
                tags_module_path.glob(private_sub_module_pattern),
            ),
        ),
    )


def get_external_tag_plugins() -> tabc.Iterator[tuple[PluginName, ModuleName]]:
    return map(attrgetter("name", "module"), entry_points(group="granular_configuration_language_20_tag"))


def get_all_tag_plugins(*, disable_plugin: tabc.Set[str]) -> tabc.Iterator[tuple[ModuleName, PluginName]]:
    builtin = PluginName("<gcl-built-in>")

    for module in get_internal_tag_plugins():
        yield module, builtin

    for name, module in get_external_tag_plugins():
        if name not in disable_plugin:
            yield module, name


def load_tags(
    *,
    disable_plugins: tabc.Set[str] = frozenset(),
    disable_tags: tabc.Set[Tag | str] = frozenset(),
) -> TagSet:
    disable_plugins |= frozenset(filter(None, map(str.strip, os.getenv("G_CONFIG_DISABLE_PLUGINS", "").split(","))))
    disable_tags |= frozenset(filter(None, map(str.strip, os.getenv("G_CONFIG_DISABLE_TAGS", "").split(","))))

    return TagSet(
        filterfalse(
            disable_tags.__contains__,
            chain.from_iterable(starmap(get_tags_in_module, get_all_tag_plugins(disable_plugin=disable_plugins))),
        )
    )



from __future__ import annotations

import collections.abc as tabc
import inspect
import itertools
import operator as op
import typing as typ
from collections import OrderedDict

from granular_configuration_language.yaml.decorators._base import TagConstructor
from granular_configuration_language.yaml.decorators._tag_tracker import is_not_lazy, is_with_ref, is_without_ref

try:
    import tabulate

    can_table = True
except ImportError:  # pragma: no cover
    tabulate = None  # type: ignore[assignment]
    can_table = False


Imode = typ.Literal["full"] | typ.Literal["reduced"] | typ.Literal[""]


class RowType(typ.TypedDict, total=False):
    plugin: str
    category: str
    tag: str
    type: str
    interpolates: Imode
    lazy: typ.Literal["NOT_LAZY"] | typ.Literal[""]
    returns: str
    handler: str


def _interpolates(tc: TagConstructor) -> Imode:
    if is_with_ref(tc.constructor):
        return "full"
    elif is_without_ref(tc.constructor):
        return "reduced"
    else:
        return ""


def _make_row(tc: TagConstructor) -> RowType:
    return RowType(
        plugin=tc.plugin,
        category=tc.category,
        tag=tc.tag,
        type=tc.friendly_type,
        interpolates=_interpolates(tc),
        lazy="NOT_LAZY" if is_not_lazy(tc.constructor) else "",
        returns=str(inspect.signature(tc.constructor).return_annotation).removeprefix("typ.").removeprefix("tabc."),
        handler=tc.constructor.__module__ + "." + tc.constructor.__name__,
    )


class AvailableBase:
    headers: tabc.Sequence[str]
    sort_keys: op.attrgetter

    def __init__(self, tags: tabc.Iterable[TagConstructor]) -> None:
        self.tags: typ.Final = tags

    def __remove_keys(self, row: RowType) -> RowType:
        keys_to_del = row.keys() - set(self.headers)

        for key in keys_to_del:
            del row[key]  # type: ignore[misc]
        return row

    def get_rows(self) -> tabc.Iterator[RowType]:
        return map(self.__remove_keys, map(_make_row, sorted(self.tags, key=self.sort_keys)))

    def table(self, *, _force_missing: bool = False) -> str:
        if tabulate and not _force_missing:
            return tabulate.tabulate(self.get_rows(), headers=OrderedDict(zip(self.headers, self.headers)))
        else:
            return """\
The "table" option requires `tabulate` to be installed.
You can use the "printing" extra to install the needed dependencies"""

    def csv(self) -> str:
        import csv
        import io

        with io.StringIO(newline="") as output:
            writer = csv.DictWriter(output, self.headers, lineterminator="\n")
            writer.writeheader()
            writer.writerows(self.get_rows())
            return output.getvalue()[:-1]


class AvailableTags(AvailableBase):
    headers = ("category", "tag", "type", "interpolates", "lazy", "returns")
    sort_keys = op.attrgetter("category", "sort_as")

    def json(self) -> str:
        import json

        def strip_tag(row: RowType) -> tuple[str, RowType]:
            tag = row.pop("tag")
            del row["category"]
            return tag, row

        category_key = op.itemgetter("category")

        def category_group(category: str, group: tabc.Iterator[RowType]) -> tuple[str, dict[str, RowType]]:
            return category, dict(map(strip_tag, group))

        return json.dumps(
            dict(itertools.starmap(category_group, itertools.groupby(self.get_rows(), key=category_key))),
            sort_keys=True,
            indent=2,
        )


class AvailablePlugins(AvailableBase):
    headers = ("plugin", "category", "tag", "handler")
    sort_keys = op.attrgetter("plugin", "category", "sort_as")

    def json(self) -> str:
        import json

        def strip_tag(row: RowType) -> tuple[str, RowType]:
            tag = row.pop("tag")
            del row["category"]
            return tag, row

        category_key = op.itemgetter("category")

        def category_group(category: str, group: tabc.Iterator[RowType]) -> tuple[str, dict[str, RowType]]:
            return category, dict(map(strip_tag, group))

        plugin_key = op.itemgetter("plugin")

        def plugin_group(plugin: str, group: tabc.Iterator[RowType]) -> tuple[str, dict[str, dict[str, RowType]]]:
            return plugin, dict(itertools.starmap(category_group, itertools.groupby(group, key=category_key)))

        return json.dumps(
            dict(itertools.starmap(plugin_group, itertools.groupby(self.get_rows(), key=plugin_key))),
            sort_keys=True,
            indent=2,
        )
